@@ -4,7 +4,7 @@ use strict;
 use warnings::register;
 
 use vars qw($VERSION %declared);
-$VERSION = '1.18';
+$VERSION = '1.21';
 
 #=======================================================================
 
@@ -45,6 +45,7 @@ sub import {
     my $constants;
     my $multiple  = ref $_[0];
     my $pkg = caller;
+    my $flush_mro;
     my $symtab;
 
     if (_CAN_PCS) {
@@ -59,15 +60,14 @@ sub import {
 	}
 	$constants = shift;
     } else {
+	unless (defined $_[0]) {
+	    require Carp;
+	    Carp::croak("Can't use undef as constant name");
+	}
 	$constants->{+shift} = undef;
     }
 
     foreach my $name ( keys %$constants ) {
-	unless (defined $name) {
-	    require Carp;
-	    Carp::croak("Can't use undef as constant name");
-	}
-
 	# Normal constant name
 	if ($name =~ $normal_constant_name and !$forbidden{$name}) {
 	    # Everything is okay
@@ -116,14 +116,23 @@ sub import {
 	    $declared{$full_name}++;
 	    if ($multiple || @_ == 1) {
 		my $scalar = $multiple ? $constants->{$name} : $_[0];
-		if ($symtab && !exists $symtab->{$name}) {
+
+		# Work around perl bug #xxxxx: Sub names (actually glob
+		# names in general) ignore the UTF8 flag. So we have to
+		# turn it off to get the "right" symbol table entry.
+		utf8::is_utf8 $name and utf8::encode $name
+		    if defined &utf8::is_utf8;
+
+		# The constant serves to optimise this entire block out on
+		# 5.8 and earlier.
+		if (_CAN_PCS && $symtab && !exists $symtab->{$name}) {
 		    # No typeglob yet, so we can use a reference as space-
 		    # efficient proxy for a constant subroutine
 		    # The check in Perl_ck_rvconst knows that inlinable
 		    # constants from cv_const_sv are read only. So we have to:
 		    Internals::SvREADONLY($scalar, 1);
 		    $symtab->{$name} = \$scalar;
-		    mro::method_changed_in($pkg);
+		    ++$flush_mro;
 		} else {
 		    *$full_name = sub () { $scalar };
 		}
@@ -135,6 +144,8 @@ sub import {
 	    }
 	}
     }
+    # Flush the cache exactly once if we make any direct symbol table changes.
+    mro::method_changed_in($pkg) if _CAN_PCS && $flush_mro;
 }
 
 1;
